/*{{{  unit.c*/
#include <string.h>
#include  "adout.h"
#include  "label.h"	/* for 'stmt_label' and 'just_written_adapt_finish' */
#include  "output.h"
#include  "tokens.h"
#include  "unit.h"
#include  "typedef.h"

/*{{{  EXPLANATION*/
#if 0
---------------------------------------------------------------
EXPLANATION
-----------
The functions in 'unit.c' handle scope units (viz. external/internal
and module procs, interface blocks and proc interface bodies, and
derived-type definitions), and set the 'scope number' for each scope unit.

Information about each scope unit is stored in a 'scope record',
a structure of type 'Unit'.  There is a linked list of records
for the units currently 'in scope'.  'this_unitp' points to the one for
the current scope unit, and each record has a pointer 'unitp->prev'
to the record of the enclosing unit.  Records are deleted
when their units are no longer in scope.

A field 'unitp->attrs' contains flags indicating the type of unit
and (if we're in a procedure or program unit) the 'program part'.
To be precise, the flags fall into 3 groups:

/*{{{  (1) UNIT_LEVEL*/
(1) UNIT_LEVEL:
---------------
EXTERNAL_LEVEL	-- at level of external progs (incl. modules)
MODULE_LEVEL	-- at level of module procs (after CONTAINS in a module)
INTERNAL_LEVEL	-- at level of internal procs (after CONTAINS in an
		   external or module proc).
INTERFACE_BODY_LEVEL,  INTERFACE_MOD_PROC_LEVEL (collectively INTERFACE_LEVEL) :
		-- in interface block
TYPEDEF_PRIV_SEQ_LEVEL,  TYPEDEF_COMPNT_LEVEL (collectively TYPEDEF_LEVEL) :
		-- in a type definition

There can be only one EXTERNAL_LEVEL record -- it's created at the start
of 'adapt' and persists throughout the compilation.

MODULE_LEVEL and INTERNAL_LEVEL units are created as a result of a
'CONTAINS' stmt -- the former if we're in a module, the latter otherwise.
There can be only one MODULE_LEVEL unit, and there should be at most one
INTERNAL_LEVEL (but more may exist as a result of an internal proc
erroneously 'CONTAIN'ing further internal procs!)  These records are
deleted at the end of the CONTAINS section (i.e. at the end of the
enclosing prog unit).

An INTERFACE_BODY_LEVEL unit is created for an INTERFACE block.  Its level
changes to INTERFACE_MOD_PROC_LEVEL if the interface block contains
'MODULE PROCEDURE' stmts, which shouldn't be followed by explicit
procedure interfaces.  This unit is deleted at the end of the interface
block.

A TYPEDEF_PRIV_SEQ_LEVEL unit is created for a TYPE definition.
At this level, the only stmts allowed are 'seq_stmt', 'access_stmt'
(to be precise, just one particular form of this, namely "PRIVATE"),
or a 'component def' stmt.  Its level changes to TYPEDEF_COMPNT_LEVEL
when the first 'component def' stmt is seen, after which 'seq_stmt'
and 'access_stmt' are disallowed.  
This unit is always at the top of the unit stack:  obviously a type def
can't contain within it further type defs, procs or interface blocks, and
if such illegal stmts appear within a type def in the program input, an
errmsg is given and the type def terminated (by 'set_part_and_label()')
before the new unit is started.  Also, this unit *must* be nested within
a proc unit -- again, that's ensured by 'set_part_and_label()'.

The first 4 levels are usually associated with a procedure or program unit
(i.e. an external/internal or module proc, or a proc interface body).
If any unit *isn't* currently associated with a proc, a special flag
is set: NOT_IN_PROC_UNIT.  For the 4 levels mentioned above, this only
happens if we're 'between' procs/prog units, i.e. between an 'END' stmt
and a following 'header' (FUNCTION, SUBROUTINE, MODULE, etc) stmt.
When a unit *is* associated with a proc/prog unit, one flag from each
of the next 2 groups is set:
/*}}}*/
/*{{{  (2) UNIT_PROC_TYPE*/
(2) UNIT_PROC_TYPE
------------------
Indicating the type of proc or prog unit, viz:

SUBROUTINE_UNIT	-- can occur at external / module / internal / interface level
FUNCTION_UNIT	--       "      "       "      "      "      "      "      "
BLOCK_DATA_UNIT	-- can only occur at EXTERNAL_LEVEL
PROGRAM_UNIT	--       "      "       "      "
MODULE_UNIT	--       "      "       "      "
/*}}}*/
/*{{{  (3) UNIT_PART*/
(3) UNIT_PART
-------------
Indicating the current 'program part' within the proc/prog unit, viz:

HDR_PART
USE_PART
IMPNONE_PART
IMPL_PART
DECL_PART
EXEC_PART
WHERE_PART

Each stmt has a 'part' flag, which indicates the program part where
the stmt can legally appear.  The 'part' flag is compared with
'this_unitp->attrs' to see if it's in an appropriate program part,
which is incremented if necessary.  This is done by function
'set_part_and_label()' (in 'label.c'), which is called after every
stmt has been parsed.
/*}}}*/

/*{{{  Labels and block constructs*/
Labels and block constructs
---------------------------
The unit records also store information about the labels & block constructs
associated with each scope unit.

******* FILL IN DETAILS HERE *************

/*}}}*/
/*{{{  Other members of the structure 'unitp'*/
Other members of the structure 'unitp'
--------------------------------------
Some other members of the structure 'unitp' are as follows:

unitp->sp	-- the symtab entry for the name of the current proc or
		   prog unit associated with this level (for external /
		   module / internal / interface levels) or the name
		   of the type currently being defined (for a type def
		   level);
unitp->first_sp	-- the first 'sp' declared at this level;
unitp->prev_sp	-- the 'sp' such that 'prev_sp->nextid' == 'first_sp'.
unitp->scope	-- the 'scope number' of this scope unit (viz. this particular
		   prog or type def).  This is stored in the 'sp->scope'
	           field of all IDs local to this scope unit.
		   There's a global variable, 'last_scope', containing the
		   last scope number used (as they're allotted in sequence).
unitp->imptab	-- an array of 'int [26]', specifying the mapping of initial
		   letters to implicit types for this scope unit.

For interface block only, the following members are used to store info
about the 'generic-spec':

unitp->generic_tk
unitp->generic_name
unitp->generic_sp
/*}}}*/
/*{{{  'Scope type'*/
'Scope type'
------------
There is a global variable 'scope_type', which is set by 'yyparse'
depending on the program context.  It determines the expected or
allowable 'scope number' of IDs in the program.  It's used by
'lookup_or_install_id()': if a 'name' is found in the symtab which
doesn't have ther right scope number (sp->scope), then it can't be the
relevant one so a new entry is installed with the appropriate 'sp->scope'.

'scope_type' can take the following values:

SC_LOCAL	-- ID must be local to this unit:
			(sp->scope == this_unitp->scope).

SC_DEF_LOCAL	-- ID can be local to an enclosing scope unit -- the most
		   closely nested decl is the relevant one.  If an ID isn't
		   in the symtab, the reference to 'name' is regarded as
		   a decl (as allowed in Fortran), and 'name' is entered
		   in symtab with default local scope:
			(sp->scope = this_unitp->scope).

SC_DEF_HOST	-- the ID *cannot* be local, but apart from that the most
		   closely nested decl is the relevant one.  If an ID
		   satisfying this condition isn't found in the symtab,
		   it's installed with the 'host' scope number, viz:
			(sp->scope = this_unitp->prev->scope).

'yyparse' sets 'scope_type' as follows:

-- For any decl of a name:	scope_type = SC_LOCAL
-- If a name is referenced (i.e. appears in any context other than a decl),
its 'scope_type' depends on the current unit type:

(i) in an INTERFACE body:	scope_type = SC_LOCAL
(ii) in a TYPE def:		scope_type = SC_DEF_HOST
(iii) otherwise:		scope_type = SC_DEF_LOCAL

'scope_type' is set in 'yyparse()' using the macros 'set_local_scope()'
and 'reset_normal_scope()'.
/*}}}*/

---------------------------------------------------------------
#endif
/*}}}*/
/*{{{  COMMENTS -- what's wrong; what's still to do!*/
/***************************************************************
Still to do
-----------
(1) Get 'process_hdr_stmt()' to set approp 'attrs' flags in 'sp', and
take corrective action if 'sp' prev defined!!

(2) Figure out what 'unitp->generic_sp /_tk /_name' (which adorn
interface blocks) are used for, and how to process them.

(3) check the setting of 'scope_type' in 'adapt.y' -- to ensure that
it's really correct at the point where IDs are looked-up (i.e. ensure
that ID isn't looked-up early because of 'lookahead').

(4) Sort out use of WHERE_PART.

(5) When done block stacks, make nec mods to all fns below,
and to 'set_part_and_label ()'.

PROBLEM WITH 'this_unitp->first_sp':
------------------------------------
	In INTERFACE & CONTAINS parts, 'sp (proc-name)' is first installed
with 'scope no' of the enclosing unit, and then a copy (new_sp) is made
with new 'scope no'.  However, if it's the first proc in the new unit,
'lookup_or_install_id()' attaches 'sp' to 'this_unitp->first_sp' of the
*new* unit, even though according to its scope no it belongs to the
enclosing unit.  It can *never* be attached to 'this_unitp->first_sp'
of the enclosing unit, even if it should be -- as that unit is not
'this_unitp' when the name is processed.
	This matters if, e.g., the first thing in main program is:
	INTERFACE
	  FUNCTION F(X)
'F' *won't* be attached to 'this_unitp->first_sp' of external unit,
but rather that of the 'interface unit'.
	What's 'this_unitp->first_sp' used for ??  Do we need it ??
***************************************************************/
/*}}}*/
/*{{{  #defines      -- only used in 'process_end_stmt()' at the moment*/
#define  str_proc_type(type)						\
			((type & SUBROUTINE_UNIT) ? str_subroutine :	\
			((type & FUNCTION_UNIT) ? str_function :	\
			((type & BLOCK_DATA_UNIT) ? str_block_data :	\
			((type & PROGRAM_UNIT) ? str_program :		\
			((type & MODULE_UNIT) ? str_module : 0)))))

#define  str_PROC_TYPE(type)						\
			((type & SUBROUTINE_UNIT) ? str_SUBROUTINE :	\
			((type & FUNCTION_UNIT) ? str_FUNCTION :	\
			((type & BLOCK_DATA_UNIT) ? str_BLOCK_DATA :	\
			((type & PROGRAM_UNIT) ? str_PROGRAM :		\
			((type & MODULE_UNIT) ? str_MODULE : 0)))))

#define  EXCL_PROC	FALSE
#define  INCL_PROC	TRUE
/*}}}*/
/*{{{  global decls  -- only used in 'process_end_stmt()' at the moment*/

char *str_subroutine = "subroutine",   *str_SUBROUTINE = "SUBROUTINE",
     *str_function   = "function",     *str_FUNCTION   = "FUNCTION",
     *str_block_data = "block data",   *str_BLOCK_DATA = "BLOCK DATA",
     *str_program    = "main program", *str_PROGRAM    = "PROGRAM",
     *str_module     = "module",       *str_MODULE     = "MODULE";
     
char  *brackets_msg = "(where [..] denotes an optional item)";

static int  last_scope = 0;	/* value of 'unitp->scope' in the */
				/* last unit opened!              */

/* definitions of 'extern's in 'unit.h' */

int   scope_type;
Unit  *this_unitp;
/*}}}*/


/*{{{  create_unit_record (level, sp) -- ok*/
create_unit_record (level, sp)
/*--------------------------*/
	int  level;
	Symbol  *sp;	/* only for TYPE definitions!  */
{
	Unit  *unitp;

	/*------------------------------------------------------*/
	/* Create a new 'unit' record, and add it to the front	*/
	/* of the 'unit' list.					*/
	/*------------------------------------------------------*/
	unitp = (Unit *) ecalloc (sizeof (Unit));
				/* all fields are initialised to 0 */
	unitp->prev = this_unitp;
	
	unitp->sp    = sp;
	unitp->attrs = level | NOT_IN_PROC_UNIT;
	init_imptab (unitp);
	unitp->prev_sp = last_sp;

	/*------------------------------------------------------*/
	/* Set a new local 'scope number' for TYPE defs (which	*/
	/* applies to all the TYPE components).			*/
	/*------------------------------------------------------*/
	if (level == TYPEDEF_PRIV_SEQ_LEVEL)
		unitp->scope = ++last_scope;

	/*------------------------------------------------------*/
	/* ... for all other unit types, copy the 'scope no'	*/
	/* from the enclosing unit, if there is one, otherwise	*/
	/* leave it at 0.					*/
	/* N.B. This 'scope no' only applies to proc names ---	*/
	/* which are visible within the enclosing unit.  After	*/
	/* reading the proc name, 'scope no' is set to a new	*/
	/* value, local to that proc, by 'process_hdr_stmt()'.	*/
	/*------------------------------------------------------*/
	else if (this_unitp)
		unitp->scope = this_unitp->scope;

	/*------------------------------------------------------*/
	/* Initialise the label information for the unit.	*/
	/* N.B. Only 'unitp->fmt_newlabel' need be explicitly	*/
	/* initialised; 'std_newlabel' & 'first_lp' start at 0.	*/
	/*------------------------------------------------------*/
	unitp->fmt_newlabel = 100000;
	
	this_unitp = unitp;
	reset_normal_scope ();	/* to that appropriate to the new unit! */
}
/*}}}*/
/*{{{  remove_unit_record () -- ok*/
remove_unit_record ()
/*-----------------*/
{
	Unit  *prev_unitp = this_unitp->prev;

	clear_labels ();	/* in case, e.g., an 'END' stmt at the    */
				/* end of a CONTAINS section is labelled. */

	free ((char *) this_unitp);
	this_unitp = prev_unitp;
	reset_normal_scope ();	/* to that appropriate to the revealed unit. */
				/* It's assumed that (this_unitp != NULL).   */
}
/*}}}*/

/*{{{  process_hdr_stmt (tk, sp, generic_tk, generic_name, generic_sp) -- ok*/
process_hdr_stmt (tk, sp, generic_tk, generic_name, generic_sp)
/*-----------------------------------------------------------*/
	int  tk;
	Symbol  *sp;		/* proc name -- may be NULL */
	int  generic_tk;	/* token for OPERATOR () or ASSIGNMENT (=) */
	char  *generic_name;	/* used if (generic_tk == TK_REL_OP) */
	Symbol  *generic_sp;	/* used for 'generic name' or TK_DERIVED_OP */
/*--------------------------------------------------------------*/
/* Calls to this function are organised as follows:		*/
/*								*/
/* switch (tk) {						*/
/* TK_EOF:							*/
/*	called alone.						*/
/*								*/
/* TK_PROGRAM : TK_MODULE : TK_BLOCK_DATA : TK_FUNCTION :	*/
/* TK_SUBROUTINE :	-- set label/part *after* unit started.	*/
/* 	process_hdr_stmt (TK_..., sp, 0,0,0);			*/
/*	set_part_and_label (HDR_PART, 0);			*/
/*								*/
/* TK_INTERFACE : TK_TYPE :					*/
/*	set_part_and_label (DECL_PART, 0);			*/
/*	if (this_unitp->attrs & IN_PROC_UNIT)			*/
/*		process_hdr_stmt (TK_..., ...);			*/
/*								*/
/* TK_CONTAINS :						*/
/*	set_part_and_label (DONT_TEST_PART, 0);			*/
/*	process_hdr_stmt (TK_CONTAINS, 0,0,0,0);		*/
/* }								*/
/*--------------------------------------------------------------*/
{
	int  this_level = this_unitp->attrs & UNIT_LEVEL;
	Symbol *new_sp;		/* for subroutine or function */
	int32  new_attr;
	
	switch (tk) {
	case TK_CONTAINS :
		/*{{{  -- ok*/
		/*--------------------------------------------------------------*/
		/* This may be invoked in any context (i.e. at any unit level,	*/
		/* inside or outside a proc), as 'set_stmt_and_part()' is	*/
		/* called with parameter 'DONT_TEST_PART' and so doesn't perform*/
		/* any tests.  The reason is that we often wish to perform	*/
		/* non-standard actions in the event of an unusual context --	*/
		/* e.g. if we're outside all prog units, we *ignore* the stmt	*/
		/* rather than start a main program.  Hence, all such tests are	*/
		/* done here.							*/
		/*--------------------------------------------------------------*/
		
		/*{{{  If not in proc a unit, give errmsg and return (i.e. ignore CONTAINS stmt)*/
		/*--------------------------------------------------------------*/
		/* If not in proc unit, give errmsg and return (i.e. ignore	*/
		/* CONTAINS stmt!)						*/
		/*--------------------------------------------------------------*/
		if (this_unitp->attrs & NOT_IN_PROC_UNIT) {
			switch (this_level) {
			case  EXTERNAL_LEVEL :
				errmsg ("\'CONTAINS\' statement not in a program unit");
				return;
		
			case  INTERNAL_LEVEL :
			case  MODULE_LEVEL :
				errmsg ("illegal statement in \'CONTAINS\' part: only procedures are allowed");
				return;
				
			case  INTERFACE_BODY_LEVEL :
			case  INTERFACE_MOD_PROC_LEVEL :
				errmsg ("illegal statement in interface block: only procedures and \'MODULE PROPCEDURE\' statements are allowed");
				return;
			
			case  TYPEDEF_PRIV_SEQ_LEVEL :
			case  TYPEDEF_COMPNT_LEVEL :
				errmsg ("illegal statement in type definition");
				return;
			}
		}
		/*}}}*/
		
		/*{{{  In a proc unit.  First, check that blocks are closed.*/
		/*--------------------------------------------------------------*/
		/* We're in a program unit.  Firstly, check that all block	*/
		/* constructs are terminated.					*/
		/*--------------------------------------------------------------*/
		clear_blocks ();
		/*}}}*/
		
		/*{{{  If we're in a MODULE, start 'MODULE_LEVEL' unit and return*/
		/*--------------------------------------------------------------*/
		/* If we're in a module, 'CONTAINS' introduces module procs	*/
		/* so create a 'MODULE_LEVEL' unit...				*/
		/*--------------------------------------------------------------*/
		if (this_unitp->attrs & MODULE_UNIT) {
			create_unit_record (MODULE_LEVEL, 0);
			return;
		}
		/*}}}*/
		
		/*{{{  --- otherwise, start 'INTERNAL_LEVEL' unit, giving errmsg if necessary*/
		/*--------------------------------------------------------------*/
		/* ... otherwise, 'CONTAINS' introduces internal procs, so	*/
		/* create an 'INTERNAL_LEVEL' unit.				*/
		/*	If we're in an internal proc or a proc interface,	*/
		/* internal procs are disallowed so give an errmsg.  Otherwise,	*/
		/* give a warning if there are no executable stmts preceding	*/
		/* the 'CONTAINS' stmt (we say 'otherwise' in order to limit	*/
		/* the number of potential messages to one!).  In any event,	*/
		/* however, we still create an 'INTERNAL_LEVEL' unit so that	*/
		/* the input can continue to be processed.			*/
		/*--------------------------------------------------------------*/
		switch (this_level) {
		case INTERNAL_LEVEL :
			errmsg ("internal procedure %q must not CONTAIN further internal procedures",
				this_unitp->sp ? this_unitp->sp->name : 0);
			break;
		
		case INTERFACE_BODY_LEVEL :
			errmsg ("interface body for procedure %q must not specify internal procedures",
				this_unitp->sp ? this_unitp->sp->name : 0);
			break;
		
		default :
			if ( !(this_unitp->attrs & EXEC_PART))
				warning ("no executable statements, so internal procedures (introduced by \'CONTAINS\') are redundant");
			break;
		}
		create_unit_record (INTERNAL_LEVEL, 0);
		return;
		
		/*}}}*/
		/*}}}*/

	case TK_FUNCTION :
	case TK_SUBROUTINE :
		/*{{{  -- ok*/
		/*--------------------------------------------------------------*/
		/* A procedure (i.e. a SUBROUTINE or FUNCTION) can occur at	*/
		/* any level of program unit (viz. external, module, internal	*/
		/* or interface block), but not in a TYPE def.			*/
		/*	If we're in the latter, close it with an errmsg.  Then,	*/
		/* if we're in a program unit, close it also with an errmsg,	*/
		/* then start this unit at the same level.			*/
		/*--------------------------------------------------------------*/
		if (this_level & TYPEDEF_LEVEL) {
			/*{{{  close TYPE def with errmsg, and reset 'this_level'*/
			/*--------------------------------------------------------------*/
			/* Close the 'TYPE def' with an errmsg.				*/
			/* 								*/
			/* N.B. The correctness of this code relies on the fact that a	*/
			/* 'TYPE def' unit is always nested inside a proc unit		*/
			/* (see the intro comments).					*/
			/*--------------------------------------------------------------*/
			missing_end_for_units_upto (0, UNIT_PROC_TYPE, EXCL_PROC);
			this_level = this_unitp->attrs & UNIT_LEVEL;
			
			/*}}}*/
		}
		
		if (this_unitp->attrs & UNIT_PROC_TYPE) {
			/*{{{  close current proc with errmsg*/
			
			if (sp  &&  (this_level == EXTERNAL_LEVEL)) {
				/*------------------------------------------------------*/
				/* Save 'sp->name', as its symtab entry will be		*/
				/* removed!  Also, since we have to re-install the	*/
				/* name, re-compute its 'len' and 'hashval'.		*/
				/*------------------------------------------------------*/
				int  len,  hashval;
				char *name;
			
				len = hashval = 0;
				for (name = sp->name;  *name;  name++) {
					len++;
					hashval = *name + (31 * hashval);
				}
				hashval = hashval % HASHSIZE;
				
				name  = (char *) emalloc (len+1);	/* +1 for '\0' */
				strcpy (name, sp->name);
			
				missing_end_for_units_upto (this_unitp, 0, INCL_PROC);
				sp = lookup_or_install_id (name, len, hashval);	/* always installed! */
				free (name);
			}
			else
				missing_end_for_units_upto (this_unitp, 0, INCL_PROC);
			/*}}}*/
		}
		
		this_unitp->sp    = sp;
		this_unitp->attrs = this_level | HDR_PART |
		                       ((tk == TK_FUNCTION) ? FUNCTION_UNIT : SUBROUTINE_UNIT);
		if (this_level != EXTERNAL_LEVEL)
			this_unitp->scope = ++last_scope;
		
		procname_msg ();	/* do this now, so 'proc-name' is o/p	*/
					/* before any errmsgs relating to it!	*/
		if (!sp)
			return;		/* an error, but nothing more to do! */
		
		if (this_level == EXTERNAL_LEVEL) {
			/*------------------------------------------------------*/
			/* 'install_global_entity ()' puts 'sp' in the		*/
			/* 'global_list' (i.e. the list of global entities) and	*/
			/* installs the appropriate flag in 'sp->attrs'.	*/
			/*	There won't be problems, either with name-clash	*/
			/* or installing 'sp->attrs', as '*sp' is newly created	*/
			/* and is the first ID in the symbol table!		*/
			/*------------------------------------------------------*/
			new_attr = ((tk == TK_FUNCTION) ? AT_RESULT_VAR : AT_SUBR);
		
			install_global_entity (sp, new_attr);
			return;
		}
		
		/*--------------------------------------------------------------*/
		/* At this point, we're dealing with a procedure interface body	*/
		/* or an internal or module procedure.				*/
		/*--------------------------------------------------------------*/
		
		/*{{{  make a new symtab entry for 'proc-name' with local scope*/
		/*-----------------------------------------------------------------
			Since we're not at the external level, this is the start of
		a procedure interface body, or an internal or module procedure.
			Make a new symtab entry, 'new_sp', for the 'proc-name',
		with the new local scope number, and set 'this_unitp->sp' to point to it.
		
		N.B. Note that there are now *2* symtab entries for 'proc-name':
		the original 'sp', whose 'scope number' is that of the enclosing
		unit, and the 'new_sp' with local scope. 
			It's necessary to make a local copy of 'proc-name' for
		several reasons:
		
		(i) it's really a local variable in the procedure interface body
			or definition;
		(ii) a 'function-name' has a different usage and thus a different
			attr flag in its definition and caller (AT_RESULT_VAR and
			AT_FUNC resp.);
		(iii) the caller may install different information for 'proc-name'
			to that installed in the definition (e.g. EXTERNAL attr,
			a specific type & shape, etc), and this should be kept separate;
		(iv) we need separate symtab entries in case the caller uses 'proc-name'
			for a different purpose than as a procedure (in which case
			an error is reported);
		(v) finally, this is consistent with what is done for INTRINSIC procs,
			where we don't wish to attach the 'intrintab' entry directly
			to the symtab as it may be altered or corrupted.
		-----------------------------------------------------------------*/
		
		this_unitp->sp = new_sp =
		        lookup_or_install_id (sp->name, strlen (sp->name), hash (sp->name));
		
			/* ... always installed, not 'looked up'! */
		
		new_sp->attrs = (tk == TK_FUNCTION) ? AT_RESULT_VAR : AT_SUBR;
		/*}}}*/
		
		/*{{{  if 'sp' is valid for the proc, set 'sp->info.sp = new_sp'*/
		/*--------------------------------------------------------------*/
		/* 'sp' is an interface body / internal / module procedure.	*/
		/* --------------------------------------------------------	*/
		/* If 'sp' correctly identifies the procedure in the caller	*/
		/* (i.e. point (iv) above doesn't apply), set:			*/
		/*								*/
		/*	sp->info.sp = new_sp					*/
		/*								*/
		/* Otherwise, report an appropriate errmsg.			*/
		/*--------------------------------------------------------------*/
		
		if ((sp->attrs & (AT_FUNC | AT_SUBR))  &&  (sp->attrs & AT_INFO)) {
			/*{{{  proc interface or definition already given.  Give errmsg & return*/
			/*--------------------------------------------------------------*/
			/* Procedure definition or interface already provided.  Give an	*/
			/* errmsg & return.						*/
			/*								*/
			/* N.B.  This branch is executed even if the previous interface	*/
			/* or definition was a subroutine and this is a function, or	*/
			/* vice versa.							*/
			/*								*/
			/* N.B. A PROBLEM:						*/
			/* ---------------						*/
			/* REALLY, 'sp' SHOULDN'T BE *USED* AS AN INTRINSIC IF IT	*/
			/* HASN'T BEEN DECLARED AS SUCH, AND IT HAS AN EXPLICIT		*/
			/* INTERFACE OR IS DEFINED AS AN INTERNAL/MODULE PROC.		*/
			/*  NEED TO LOOKAHEAD TO ESTABLISH THIS.  FIX THIS UP LATER!	*/
			/*	For the moment, we don't give an errmsg for the case	*/
			/* where 'sp' was used as an intrinsic but wasn't declared as	*/
			/* such.  We just quitely forget about it!			*/
			/*--------------------------------------------------------------*/
			
			if ((sp->info.sp->attrs & AT_INTRINSIC)  &&  !(sp->attrs & AT_INTRINSIC))
				;		/* Keep quiet!  See above! */
				
			/*-----------------------------------------------*/
			/* an internal or module procedure definition... */
			/*-----------------------------------------------*/
			else if (this_level & (MODULE_LEVEL | INTERNAL_LEVEL)) {
				begin_errmsg ("%s procedure %q ",
				         (this_level == MODULE_LEVEL) ? "module" : "internal",
				         sp->name);
			
				if (sp->attrs & AT_INTRINSIC)
					end_errmsg ("has been declared INTRINSIC");
				else
					end_errmsg ("is already defined or has an interface body (illegally) specified");
			}
			/*------------------------------*/
			/* ... or an interface body ...	*/
			/*------------------------------*/
			else if (sp->attrs & AT_INTRINSIC)
				errmsg ("must not specify an interface body for an intrinsic procedure: %q",
				         sp->name);
			else
				errmsg ("procedure %q already has an explicit interface",
				         sp->name);
			
			return;
			/*}}}*/
		}
		
		new_attr = ((tk == TK_FUNCTION) ? AT_FUNC : AT_SUBR);
		
		if (! install_attr (sp, new_attr))
			return;		/* 'sp' isn't a proc, or is the wrong kind of proc! */
		
		if (this_level & INTERFACE_LEVEL) {
			/*-----------------------*/
			/* a proc interface body */
			/*-----------------------*/
			/*{{{  if it isn't the host or a dummy proc, install it in 'global_list'*/
			if (sp == this_unitp->prev->sp) {  	/* N.B.'sp' is in host scope unit! */
				/*------------------------------------------------------*/
				/* 'sp' has the same name as its (direct) host, i.e.	*/
				/* the host procedure specifies its own interface.	*/
				/*  Give an errmsg and return.				*/
				/*							*/
				/* N.B.:						*/
				/* (1)  'sp' must be a subroutine or a function with a	*/
				/* separate result variable.  If it were a function	*/
				/* *without* a separate result variable, 'install_attr	*/
				/* (sp, AT_FUNC)' would have failed as 'sp' can't have	*/
				/* AT_FUNC set, since it already has AT_RESULT_VAR set!	*/
				/* (2)  This is done before 'install_global_entity ()',	*/
				/* otherwise we could install an internal or module	*/
				/* procedure in the 'global_list' !			*/
				/*------------------------------------------------------*/
				errmsg ("must not specify an interface body for the host procedure: %q",
				         sp->name);
				return;
			}
			if ( !(sp->attrs & AT_DUMMY)) {
				if ( !install_global_entity (sp, new_attr))
					return;
			}
			/*}}}*/
		}
		else {
			/*---------------------------*/
			/* a module or internal proc */
			/*---------------------------*/
			/*{{{  check it isn't the host proc, or a dummy arg, or declared EXTERNAL*/
			if (sp == this_unitp->prev->sp) {  	/* N.B.'sp' is in host scope unit! */
				/*--------------------------------------------------------------*/
				/* 'sp' has the same name as its host, i.e. the host procedure	*/
				/* CONTAINs itself.  Give an errmsg and return.			*/
				/*								*/
				/* N.B.  'sp' must be an *internal subroutine*.  If it were a	*/
				/* function, 'install_attr (sp, AT_FUNC)' would have failed	*/
				/* because 'sp' can't have AT_FUNC set, as it has AT_RESULT_VAR	*/
				/* set!.  It can't be a *module* subroutine, for then		*/
				/* 'install_attr (sp, AT_SUBR)' would fail because 'sp' is a	*/
				/* *module* name, not a subroutine name!			*/
				/*--------------------------------------------------------------*/
				errmsg ("internal subroutine %q has the same name as its host", sp->name);
				return;
			}
			else if (sp->attrs & AT_DUMMY) {
				errmsg ("dummy argument %q cannot be an internal procedure", sp->name);
				return;
			}
			else if (sp->attrs & AT_EXTERNAL) {
				errmsg ("%s procedure %q has been declared EXTERNAL",
				         (this_level == MODULE_LEVEL) ? "module" : "internal",
				         sp->name);
				return;
			}
			/*}}}*/
		}
		
		/*-----------------------------*/
		/* OK!  Link 'new_sp' to 'sp'. */
		/*-----------------------------*/
		sp->info.sp = new_sp;
		sp->attrs |= AT_INFO;		/* we now know this is OK */
		return;
		/*}}}*/
		/*}}}*/

	case TK_BLOCK_DATA :
	case TK_MODULE :
	case TK_PROGRAM :
	case TK_EOF :
		/*{{{  -- ok*/
		/*--------------------------------------------------------------*/
		/*	A BLOCK DATA, MODULE or PROGRAM unit can only occur at	*/
		/* 'EXTERNAL_LEVEL'.  Thus, we first close all current units,	*/
		/* if any, with errmsgs about "missing END stmt", before	*/
		/* starting this unit.						*/
		/*	If (tk == TK_EOF), the end-of-file has occurred.	*/
		/* In that case, the same checks and remedial actions are	*/
		/* performed as above, but then this fuction returns without	*/
		/* starting a unit (obviously!).				*/
		/*								*/
		/* N.B.: 'sp' may be NULL!					*/
		/*--------------------------------------------------------------*/
		
		if ( !( (this_unitp->attrs & NOT_IN_PROC_UNIT)  &&
		        (this_unitp->attrs & EXTERNAL_LEVEL) ) ) {
			/*{{{  close current prog unit with errmsg*/
			
			if (sp) {
				/*------------------------------------------------------*/
				/* Save 'sp->name', as its symtab entry will be		*/
				/* removed!  Also, since we have to re-install		*/
				/* 'sp->name', re-compute its 'len' and 'hashval'.	*/
				/*------------------------------------------------------*/
				int  len,  hashval;
				char *p;
			
				len = hashval = 0;
				for (p = sp->name;  *p;  p++) {
					len++;
					hashval = *p + (31 * hashval);
				}
				hashval = hashval % HASHSIZE;
				
				p  = (char *) emalloc (len+1);	/* +1 for '\0' */
				strcpy (p, sp->name);
			
				missing_end_for_units_upto (0, EXTERNAL_LEVEL, INCL_PROC);
				sp = lookup_or_install_id (p, len, hashval);	/* always installed! */
				free (p);
			}
			else
				missing_end_for_units_upto (0, EXTERNAL_LEVEL, INCL_PROC);
			
			
			/*}}}*/
		}
		
		if (tk == TK_EOF)
			return;
		
		this_unitp->sp    = sp;
		this_unitp->attrs = EXTERNAL_LEVEL | HDR_PART |
		                                    ((tk == TK_PROGRAM) ? PROGRAM_UNIT :
		                                    ((tk == TK_MODULE)  ? MODULE_UNIT :
		                                                          BLOCK_DATA_UNIT));
		procname_msg ();
		
		if (sp)
			/*------------------------------------------------------*/
			/* 'install_global_entity ()' puts 'sp' in the		*/
			/* 'global_list' (i.e. the list of global entities) and	*/
			/* installs the appropriate flag in 'sp->attrs'.	*/
			/*	There won't be problems, either with name-clash	*/
			/* or installing 'sp->attrs', as '*sp' is newly created	*/
			/* and is the first ID in the symbol table!		*/
			/*------------------------------------------------------*/
			install_global_entity (sp,  ((tk == TK_PROGRAM) ? AT_PROGRAM :
				                    ((tk == TK_MODULE)  ? AT_MODULE :
			                                                  AT_BLOCK_DATA)) );
		
		return;
		/*}}}*/

	case TK_INTERFACE :
		/*{{{  -- ok*/
		/*--------------------------------------------------------------*/
		/* Note that the name (if any) in an INTERFACE stmt is stored	*/
		/* in:								*/
		/*								*/
		/*	'unitp->generic_name'					*/
		/*								*/
		/* rather than in 'unitp->sp->name' (which is reserved for the	*/
		/* names of procedure bodies contained within the interface	*/
		/* block).  Also, if the interface describes a generic operator	*/
		/* or assignment ('INTERFACE OPERATOR (..)' or 'ASSIGNMENT (=)')*/
		/* then the operator's token is stored in:			*/
		/*								*/
		/*	'unitp->generic_tk'					*/
		/*								*/
		/* (which is unused otherwise!)					*/
		/*								*/
		/* N.B. 'yyparse()' only calls this function if the INTERFACE	*/
		/* stmt appears in a suitable context (viz. within a proc --	*/
		/* see intro comments), so no further checks are necessary.	*/
		/*--------------------------------------------------------------*/
		
		create_unit_record (INTERFACE_BODY_LEVEL, 0);
		this_unitp->generic_tk   = generic_tk;
		this_unitp->generic_name = generic_name;
		this_unitp->generic_sp   = generic_sp;
		
		return;
		/*}}}*/

	case TK_TYPE :
		/*{{{  -- ok*/
		/*--------------------------------------------------------------*/
		/* N.B. 'yyparse()' only calls this function if the TYPE stmt	*/
		/* appears in a suitable context (viz. within a proc -- see	*/
		/* intro comments), so no further checks are necessary.		*/
		/*--------------------------------------------------------------*/
		
		create_unit_record (TYPEDEF_PRIV_SEQ_LEVEL, sp);
		return;
		
		/*}}}*/
	}
}
/*
-- could clean-up symbols, labels, etc, here (just to make sure no
crap's been left by illegal stmts with DONT_TEST_PART!! )
-- !! NO !!  since then it will remove label on hdr stmt itself -- as well
as the name of the new unit!
*/
/*}}}*/
/*{{{  process_end_stmt (tk, sp, syntax_error) -- ok*/
process_end_stmt (tk, sp, syntax_error)
/*-----------------------------------*/
	int  tk;
	Symbol  *sp;
	bool  syntax_error;
/*--------------------------------------------------------------*/
/* This function also calls 'set_part_and_label ()'.  This is	*/
/* because a stmt label on an 'END' stmt obviously belongs to	*/
/* the program unit to which the 'END' relates, which, because	*/
/* of program errors, may not be the current one.  In that case,*/
/* this function closes intermediate units *before* calling	*/
/* 'set_part_and_label()', and *then* closes the unit to which	*/
/* 'END' relates.						*/
/*--------------------------------------------------------------*/
{
	int    this_proc_type = this_unitp->attrs & UNIT_PROC_TYPE,
	       this_level     = this_unitp->attrs & UNIT_LEVEL,
	       end_proc_type;
	char  *this_name = this_unitp->sp ? this_unitp->sp->name : 0;
	Unit  *unitp;

	if (syntax_error) {
		/*{{{  add to "syntax error" msg, saying how stmt interpreted.*/
		/*--------------------------------------------------------------*/
		/* Add text to the end of the "syntax error" message produced	*/
		/* by 'yyerror ()', saying, e.g.:				*/
		/*								*/
		/*	".  (Statement interpreted as "END FUNCTION F")"	*/
		/*								*/
		/* to indicate that the stmt is still interpreted, and how!	*/
		/*--------------------------------------------------------------*/
		cont_errmsg (".  (Statement interpreted as \'END");
		switch (tk) {
		case  TK_END_BLOCK_DATA : cont_errmsg (" BLOCK DATA"); break;
		case  TK_END_FUNCTION :   cont_errmsg (" FUNCTION");   break;
		case  TK_END_INTERFACE :  cont_errmsg (" INTERFACE");  break;
		case  TK_END_MODULE :     cont_errmsg (" MODULE");     break;
		case  TK_END_PROGRAM :    cont_errmsg (" PROGRAM");    break;
		case  TK_END_SUBROUTINE : cont_errmsg (" SUBROUTINE"); break;
		case  TK_END_TYPE :       cont_errmsg (" TYPE");       break;
		}
		if (sp) cont_errmsg (" %s", sp->name);
		end_errmsg ("\')");
		/*}}}*/
	}

	/*{{{  return if not in any program unit!!*/
	/*------------------------------------*/
	/* Return if not in any program unit! */
	/*------------------------------------*/
	if (!this_proc_type  &&  (this_level == EXTERNAL_LEVEL)) {
		begin_errmsg ("\'END");
		switch (tk) {
		case  TK_END_BLOCK_DATA : cont_errmsg (" BLOCK DATA"); break;
		case  TK_END_FUNCTION :   cont_errmsg (" FUNCTION");   break;
		case  TK_END_INTERFACE :  cont_errmsg (" INTERFACE");  break;
		case  TK_END_MODULE :     cont_errmsg (" MODULE");     break;
		case  TK_END_PROGRAM :    cont_errmsg (" PROGRAM");    break;
		case  TK_END_SUBROUTINE : cont_errmsg (" SUBROUTINE"); break;
		case  TK_END_TYPE :       cont_errmsg (" TYPE");       break;
		}
		end_errmsg ("\' statement not in a program unit");
	
		/*------------------------------------------------------*/
		/* the following is done because this is the one	*/
		/* circumstance when 'set_part_and_label()' isn't called*/
		/* -- normally that would perform following actions!	*/
		/*------------------------------------------------------*/
		if (this_unitp->first_sp)	/* remove any symbols */
			clear_symtab ();
		if (this_unitp->first_lp)	/* remove any stmt label */
			clear_labels ();
		return;
	}
	/*}}}*/

	switch (tk) {
	case  TK_END :
		/*{{{  simple 'END' stmt: close current unit *whatever* its type!*/
		/*--------------------------------------------------------------*/
		/* A simple 'END' stmt.  This closes the current unit		*/
		/* *whatever* its type, i.e. even if it's a ...			*/
		/*	-- module or internal procedure   (which requires	*/
		/*			'END FUNCTION' or 'END SUBROUTINE'),	*/
		/*	-- derived-type unit  (which requires 'END TYPE'),	*/
		/*	-- interface unit     (requiring 'END INTERFACE').	*/
		/*--------------------------------------------------------------*/
		if (this_proc_type) {
			/*----------------------------------------------*/
			/* report error for module / internal procedure */
			/*----------------------------------------------*/
			if (this_level & (MODULE_LEVEL | INTERNAL_LEVEL))
				errmsg ("%s %s %q must be terminated by \'END %s [%s]\' %s",
					 (this_level == MODULE_LEVEL) ? "module" : "internal",
					 str_proc_type (this_proc_type),  this_name,
					 str_PROC_TYPE (this_proc_type),  this_name,
					 brackets_msg);
		}
		else if (this_level & INTERFACE_LEVEL) {
			/*--------------------------------*/
			/* terminating an INTERFACE block */
			/*--------------------------------*/
			errmsg ("interface block must be terminated by \'END INTERFACE\'");
		
			tk = 0;		/* for 'set_part_and_label()'. We don't	*/
					/* pass-in 'TK_END', as that means the	*/
					/* stmt is a valid target for a 'GOTO'!	*/
		}
		else if (this_level & TYPEDEF_LEVEL) {
			/*-------------------------------*/
			/* terminating a TYPE definition */
			/*-------------------------------*/
			errmsg ("type definition %q must be terminated by \'END TYPE [%s]\' %s",
				 this_name,  this_name,  brackets_msg);
		
			tk = 0;		/* for 'set_part_and_label()'. We don't	*/
					/* pass-in 'TK_END', as that means the	*/
					/* stmt is a valid target for a 'GOTO'!	*/
		}
		else {
			/*------------------------------------------------------*/
			/* terminating a proc at the end of a CONTAINS section	*/
			/*------------------------------------------------------*/
			missing_end_for_units_upto (0, IN_PROC_UNIT, EXCL_PROC);
		}
		
		set_part_and_label (DONT_TEST_PART, tk);  /* 'DONT_TEST_PART' avoids repeating */
							  /* errmsgs if we're in a WHERE! */
		close_current_unit ();
		return;
		/*}}}*/

	case  TK_END_FUNCTION :
	case  TK_END_SUBROUTINE :
		/*{{{  */
		/*--------------------------------------------------------------*/
		/* An 'END FUNCTION' or 'END SUBROUTINE' stmt.			*/
		/* -------------------------------------------			*/
		/* This closes the *current* proc (viz. FUNCTION, SUBROUTINE,	*/
		/* PROGRAM, BLOCK DATA or MODULE), whatever its type and name,	*/
		/* unless:							*/
		/*  (i) a 'name' is provided on the 'END' stmt, and		*/
		/* (ii) neither name nor unit-type match the current procedure	*/
		/*      but *both* match an enclosing procedure,		*/
		/* in which case the latter is closed.				*/
		/*--------------------------------------------------------------*/
		
		end_proc_type = ((tk == TK_END_FUNCTION) ? FUNCTION_UNIT : SUBROUTINE_UNIT);
		
		/*{{{  close units until we reach the current proc unit*/
		/*--------------------------------------------------------------*/
		/* Close units in the unit stack until we reach the current proc*/
		/* (i.e. the first unit that has a 'UNIT_PROC_TYPE' flag set).	*/
		/*	To be more precise, if we're currently in a TYPE defn,	*/
		/* or an INTERFACE block but *not* within a proc therein,	*/
		/* then close it!  Also, if we're in a 'CONTAINS' section,	*/
		/* then remove the current unit record to expose the record	*/
		/* for the enclosing proc (which effectively 'ends' the		*/
		/* 'CONTAINS' section!)						*/
		/*--------------------------------------------------------------*/
		
		missing_end_for_units_upto (0, IN_PROC_UNIT, EXCL_PROC);
		/*}}}*/
		
		/*{{{  if 'end_proc_type' & 'end_name' match deeper unit but not this one, close intervening units.*/
		/*--------------------------------------------------------------*/
		/* If neither 'end_proc_type' nor 'end_name' match the current	*/
		/* unit, and there is a proc deeper in the unit stack that	*/
		/* matches 'end_proc_type' *and* 'end_name', then close inter-	*/
		/* vening units to make it the current unit.			*/
		/*--------------------------------------------------------------*/
		
		if ( !(end_proc_type & this_unitp->attrs)  &&  sp  &&  (sp != this_unitp->sp)) {
		
			for (unitp = this_unitp->prev;  unitp;  unitp = unitp->prev) {
				if ((end_proc_type & unitp->attrs)  &&  (sp == unitp->sp))
					break;		/* found! */
			}
			if (unitp)
				missing_end_for_units_upto (unitp, 0, EXCL_PROC);
							/* now 'this_unitp' = 'unitp' */
		}
		
		
		/*}}}*/
		
		/*{{{  close current unit, with errmsg if 'type' or 'name' don't match*/
		
		this_proc_type = this_unitp->attrs & UNIT_PROC_TYPE;
					/* re-evaluate as 'this_unitp' may have changed! */
		
		
		if (sp  &&  sp != this_unitp->sp)		/* name mismatch */
			errmsg ("%s %q terminated by \'END %s %s\'", 
				 str_proc_type (this_proc_type),
				 this_unitp->sp ? this_unitp->sp->name : 0,
				 str_PROC_TYPE (end_proc_type),  sp->name);
		
		else if (end_proc_type != this_proc_type)	/* unit type mismatch */
			errmsg ("%s %q terminated by \'END %s\' statement", 
				 str_proc_type (this_proc_type),
				 this_unitp->sp ? this_unitp->sp->name : 0,
				 str_PROC_TYPE (end_proc_type));
		
		set_part_and_label (DONT_TEST_PART, TK_END);
						/* 'DONT_TEST_PART' avoids repeating */
						/* errmsgs if we're in a WHERE! */
		close_current_unit ();
		
		/*}}}*/
		
		return;
		/*}}}*/

	case  TK_END_BLOCK_DATA :
	case  TK_END_MODULE :
	case  TK_END_PROGRAM :
		/*{{{  */
		/*--------------------------------------------------------------*/
		/* An 'END BLOCK DATA', 'END MODULE' or 'END PROGRAM' stmt.	*/
		/* --------------------------------------------------------	*/
		/* This closes the *current* proc (viz. FUNCTION, SUBROUTINE,	*/
		/* PROGRAM, BLOCK DATA or MODULE), whatever its type and name,	*/
		/* unless:							*/
		/*	neither the 'end_proc_type' nor the 'end_name' (if one	*/
		/*	is supplied) match the current unit, but 'end_proc_type'*/
		/*	matches the enclosing external program unit (regardless	*/
		/*	of whether 'end_name' matches, if one is supplied).	*/
		/*								*/
		/* in which case the latter is closed.				*/
		/*--------------------------------------------------------------*/
		
		end_proc_type = ((tk == TK_END_BLOCK_DATA) ? BLOCK_DATA_UNIT :
		                ((tk == TK_END_MODULE)     ? MODULE_UNIT : PROGRAM_UNIT));
		
		/*{{{  close units until we reach the current proc unit*/
		/*--------------------------------------------------------------*/
		/* Close units in the unit stack until we reach the current proc*/
		/* (i.e. the first unit that has a 'UNIT_PROC_TYPE' flag set).	*/
		/*	To be more precise, if we're currently in a TYPE defn,	*/
		/* or an INTERFACE block but *not* within a proc therein,	*/
		/* then close it!  Also, if we're in a 'CONTAINS' section,	*/
		/* then remove the current unit record to expose the record	*/
		/* for the enclosing proc (which effectively 'ends' the		*/
		/* 'CONTAINS' section!)						*/
		/*--------------------------------------------------------------*/
		
		missing_end_for_units_upto (0, IN_PROC_UNIT, EXCL_PROC);
		/*}}}*/
		
		/*{{{  if 'end_proc_type' matches deeper unit but not this one, close intervening units.*/
		/*--------------------------------------------------------------*/
		/* If neither the 'end_proc_type' nor the 'end_name' (if one is	*/
		/* supplied) match the current unit, but 'end_proc_type'	*/
		/* matches the enclosing external program unit -- regardless of	*/
		/* whether 'end_name' matches -- then close intervening units	*/
		/* to make it the current unit.					*/
		/*--------------------------------------------------------------*/
		
		if ( !(end_proc_type & this_unitp->attrs || (sp  &&  sp == this_unitp->sp)) ) {
		
			for (unitp = this_unitp->prev;  unitp;  unitp = unitp->prev) {
				if (end_proc_type & unitp->attrs)
					break;		/* found! */
			}
			if (unitp)
				missing_end_for_units_upto (unitp, 0, EXCL_PROC);
							/* now 'this_unitp' = 'unitp' */
		}
		
		
		/*}}}*/
		
		/*{{{  close current unit, with errmsg if 'type' or 'name' don't match*/
		
		this_proc_type = this_unitp->attrs & UNIT_PROC_TYPE;
					/* re-evaluate as 'this_unitp' may have changed! */
		
		
		if (sp  &&  sp != this_unitp->sp)		/* name mismatch */
			errmsg ("%s %q terminated by \'END %s %s\'", 
				 str_proc_type (this_proc_type),
				 this_unitp->sp ? this_unitp->sp->name : 0,
				 str_PROC_TYPE (end_proc_type),  sp->name);
		
		else if (end_proc_type != this_proc_type)	/* unit type mismatch */
			errmsg ("%s %q terminated by \'END %s\' statement", 
				 str_proc_type (this_proc_type),
				 this_unitp->sp ? this_unitp->sp->name : 0,
				 str_PROC_TYPE (end_proc_type));
		
		set_part_and_label (DONT_TEST_PART, TK_END);
						/* 'DONT_TEST_PART' avoids repeating */
						/* errmsgs if we're in a WHERE! */
		close_current_unit ();
		/*}}}*/
		
		return;
		
		/*}}}*/

	case  TK_END_INTERFACE :
		/*{{{  */
		/*--------------------------------------------------------------*/
		/* 'END INTERFACE' stmt.					*/
		/* ---------------------					*/
		/* If we're in an interface block, close it (as well as any un-	*/
		/* terminated units contained therein, with suitable errmsgs);	*/
		/* otherwise give an errmsg and ignore the 'END INTERFACE' stmt.*/
		/*--------------------------------------------------------------*/
		
		for (unitp = this_unitp;  unitp;  unitp = unitp->prev) {
			if (unitp->attrs & INTERFACE_LEVEL)
				break;		/* found! */
		}
		
		if (unitp) {
			missing_end_for_units_upto (unitp, 0, INCL_PROC);
							/* now 'this_unitp' = 'unitp' */
			set_part_and_label (DONT_TEST_PART, 0);
			close_current_unit ();
		}
		else {
			errmsg ("\'END INTERFACE\' statement not in an interface block");
			set_part_and_label (DONT_TEST_PART, 0);
					/* still want to install stmt label, if any! */
		}
			
		return;
		
		/*}}}*/

	case  TK_END_TYPE :
		/*{{{  */
		/*--------------------------------------------------------------*/
		/* 'END TYPE' stmt.						*/
		/* ----------------						*/
		/* If we're currently in a TYPE defn, this stmt ends it,	*/
		/* otherwise an errmsg is given and the stmt is ignored.	*/
		/*--------------------------------------------------------------*/
		
		set_part_and_label (DONT_TEST_PART, 0);
		
		if (this_level & TYPEDEF_LEVEL) {
			if (sp  &&  sp != this_unitp->sp)	/* name mismatch */
				errmsg ("type definition %q terminated by \'END TYPE %s\'", 
					 this_unitp->sp ? this_unitp->sp->name : "",
					 sp->name);
		
			if (this_level & TYPEDEF_PRIV_SEQ_LEVEL)
				errmsg ("derived type %q has no components declared",
				         this_unitp->sp->name);
			
			close_current_unit ();
		}
		else
			errmsg ("\'END TYPE\' statement not in a derived-type definition");
		
		return;
		
		/*}}}*/
	}
}
/*}}}*/

/*{{{  close_current_unit () -- ok*/
close_current_unit ()
/*-----------------*/
/*----------------------------------------------------------------
	This function is *always* called to end the current 'scope
unit', whether normally or abnormally.  It gives no errmsgs for
abnormal termination -- that's done by the calling function;
indeed, this function is unaware of whether the termination is
normal or not.  It performs all the necessary checks (e.g. are
blocks terminated and labels set?) and actions (e.g. detaching
or removing the local symtab, and re-initialising the 'imptab').
	For subprograms, this just requires re-initialising
certain fields in the 'unit record', without removing the record
itself.
	For INTERFACE blocks (not in a proc unit) and TYPE defs,
the record is removed from the 'unit stack' and 'this_unitp' is
set to point to the next record in the stack (i.e. that of the
enclosing subprogram).
	This function is never called with:

	(this_unitp->attrs & NOT_IN_PROC_UNIT)  &&
	(this_unitp->attrs & (MODULE_LEVEL | INTERNAL_LEVEL))

(i.e. to close a subprogram at the end of a 'CONTAINS' part).
This requires two actions to be performed:  the current record
must be removed, and the subprogram unit at the next unit level
must be closed.  The first action is always performed by calling
'missing_end_for_units_upto ()', so this function only performs
the second!
----------------------------------------------------------------*/
{
	if (this_unitp->attrs & UNIT_PROC_TYPE) {
		/*{{{  'END' a program unit or subprogram*/
		/*----------------------------------------------------------------
		ENDing a program unit or subprogram.
		------------------------------------
		This involves zeroing the field 'this_unitp->sp'.
			'this_unitp->attrs' has its UNIT_PROC_TYPE and UNIT_PART
		flags unset, but the UNIT_LEVEL flag is unchanged so that we remain
		at the same 'level' (external, internal, module or interface).
			'scope' & 'imptab' are zeroed if the level is external,
		or copied from the enclosing unit otherwise.
			The fact that some info is preserved explains why the
		unit record, 'this_unitp', is not simply removed!
		----------------------------------------------------------------*/
		int  this_level = this_unitp->attrs & UNIT_LEVEL;
		
		switch (this_level) {
		case  EXTERNAL_LEVEL :
			last_scope = 0;		/* restart the 'scope' numbering. */
						/* N.B. 'this_unitp->scope' is 0.  */
			if ((this_unitp->attrs & UNIT_PART) < EXEC_PART) {
				/*----------------------------------------------*/
				/* no executable stmts, so no o/p yet.  O/p the	*/
				/* decls...					*/
				/*----------------------------------------------*/
				process_decls ();
				write_decls ();			/* only for ADAPT */
			}
			if (! just_written_adapt_finish)
				write_adapt_finish ();		/* only for ADAPT */
			reduce_indent;		/* OK if no header, as won't go to col < 6 */
			stmt ("END");
			write_ida_info ();			/* only for IDA */
			cleanup ();
			break;
		
		default :
			if (this_unitp->attrs & INTERFACE_LEVEL)
				process_proc_interface ();	/* fill in all poss info */
		
			/*{{{  if this is a func whose type was declared in enclosing unit, check it*/
			/*--------------------------------------------------------------*/
			/* If this is a function whose data type has been declared in	*/
			/* the enclosing unit, check that it's declared the same within	*/
			/* this interface or definition.				*/
			/*--------------------------------------------------------------*/
			{
				char  *proc_name = this_unitp->sp->name;
				Symbol  *sp;
			
				/*-------------------------------------------------*/
				/* Lookup corresponding 'sp' in the enclosing unit */
				/*-------------------------------------------------*/
				scope_type = SC_DEF_HOST;	/* to find 'proc_name' in enclosing unit */
			
				sp = lookup_or_install_id (proc_name, strlen (proc_name), hash (proc_name));
								/* always 'looked-up', not installed! */
				reset_normal_scope ();
			
				/*------------------------------------------------------*/
				/* Check that previously declared type, if any, matches	*/
				/* that in this interface or definition.		*/
				/*							*/
				/* N.B. The first 2 conditions in the following 'if'	*/
				/* check that 'proc-name' in the enclosing unit *has*	*/
				/* an interface or definition  (in case it erroneously	*/
				/* refers to a different object), and the last condition*/
				/* checks this interface / definition is it  (in case	*/
				/* there are multiple interface bodies or definitions	*/
				/* for the same name!)					*/
				/*------------------------------------------------------*/
				if ((sp->attrs & AT_FUNC)  &&  (sp->attrs & AT_INFO)  &&
				    (sp->info.sp == this_unitp->sp)) {
			
					int32  old_type = (sp->type & TY_TYPE),
					       new_type = (this_unitp->sp->type & TY_TYPE);
			
					if (old_type  &&  new_type  &&  (old_type != new_type)) {
						earlier_error = TRUE;
						errmsg ("type previously declared for function %q disagrees with its %s",
						          proc_name,
						         (this_level & INTERFACE_LEVEL) ? "interface body" : "definition");
						earlier_error = FALSE;
			
						clear_flags (sp->type, TY_TYPE);
					}
				}
			}
			/*}}}*/
		
			detach_symtab ();	/* detach local IDs from 'symtab[]'... */
			clear_blocks ();
			earlier_error = TRUE;
			clear_labels ();
			earlier_error = FALSE;
			this_unitp->scope = this_unitp->prev->scope;
						/* ... then inherit the host scope no. */
			break;
		}
		
		this_unitp->sp = 0;		/* but don't 'free' the symtab entry! */
		
		init_imptab (this_unitp);	/* to that of enclosing unit, or '0' if external */
		
		this_unitp->attrs = this_level + NOT_IN_PROC_UNIT;
						/* Do this *after* 'clear_labels()', as       */
						/* errmsgs aren't given if not in a proc unit.*/
		
		/*}}}*/
	}
	else if (this_unitp->attrs & TYPEDEF_LEVEL) {
		/*-------------------------*/
		/* 'END' a type definition */
		/*-------------------------*/
		detach_symtab ();	/* detach component names from 'symtab[]' */
		remove_unit_record ();
	}
	else
		/*-----------------------------------------------------*/
		/* 'END' an interface block (*not* within a proc body) */
		/*-----------------------------------------------------*/
		remove_unit_record ();
		
	return;
}
/*}}}*/
/*{{{  missing_end_for_units_upto (new_unitp, unit_attr, incl_proc) -- ok*/
missing_end_for_units_upto (new_unitp, unit_attr, incl_proc)
/*--------------------------------------------------------*/
	Unit  *new_unitp;
	int  unit_attr;
	bool  incl_proc;	/* if TRUE, close proc at the given level */
/*----------------------------------------------------------------
	This is called with either 'new_unitp' or 'unit_attr'
non-zero.  It's purpose is to close each unit, starting from
the current one, 'this_unitp', up to but excluding the unit
'*new_unitp' (if new_unitp != 0) or the first unit with
'unit_attr' set (if unit_attr != 0).  The chosen unit becomes
the new current unit, 'this_initp'.
	For each unit that is closed, an error is reported
(if appropriate) along the lines of:

	"missing END stmt for function %q"

If this ends a CONTAINS section, then no errmsg is required.

N.B.  We assume that there actually *is* a unit record with the
given address '*new_unitp' or flag 'unit_attr'.
	If (new_unitp != 0) is *not* a 'unit record', the function
will crash when it de-references (this_unitp == 0).
	If (new_unitp == 0 && unit_attr != 0), but 'unit_attr'
is *not* set in any unit record, then 'this_unitp' will be 0 on
return.  This violates the assumption made throughout this program
that 'this_unitp' is never NULL!!
----------------------------------------------------------------*/
/***************************************************
IF (stmt_lp), WE COULD IN FACT DISASSOCIATE LABEL WITH THIS UNIT,
AND RE-INSTALL IT IN THE NEW UNIT THAT'S EXPOSED.   (CURRENTLY WE
ALREADY DO THIS FOR CLOSING 'CONTAINS' PART, WHERE IT'S ESSENTIAL!)

WOULDN'T NEED TO REMOVE 'stmt_lp' FROM THE LABEL LIST (??), JUST
REMOVE ITS 'lineno' (??)
*****************************/
{
	int  proc;
	
	while ((this_unitp != new_unitp) && !(this_unitp->attrs & unit_attr)) {
		
		/*{{{  close unit & remove record, giving errmsg if necessary*/
		/*--------------------------------------------------------------*/
		/* Give errmsg about "missing 'END' stmt" if necessary, viz:	*/
		/*--------------------------------------------------------------*/
		
		if (proc = this_unitp->attrs & UNIT_PROC_TYPE) {
			/*------------------------------------------------------*/
			/* in a procedure					*/
			/* --------------					*/
			/* N.B. This can only be a FUNCTION or SUBROUTINE,	*/
			/* as other types of subprogram only occur at the	*/
			/* 'external' level, which isn't closed here!		*/
			/*------------------------------------------------------*/
			errmsg ("missing 'END' statement for %s %q",
				 str_proc_type (proc),
				 this_unitp->sp ? this_unitp->sp->name : 0);
			close_current_unit ();
		}
		
		if (this_unitp->attrs & INTERFACE_LEVEL) {
			/*-----------------------*/
			/* in an INTERFACE block */
			/*-----------------------*/
			errmsg ("missing 'END INTERFACE' statement");
			close_current_unit ();
		}
		else if (this_unitp->attrs & TYPEDEF_LEVEL) {
			/*----------------------*/
			/* in a TYPE defn.	*/
			/*----------------------*/
			errmsg ("missing 'END TYPE' statement for derived-type definition %q",
				 this_unitp->sp ? this_unitp->sp->name : 0);
			close_current_unit ();
		}
		else {
			/*------------------------------------------------------*/
			/* At the end of a CONTAINS part.  No errmsg is		*/
			/* necessary, as it isn't terminated by an 'END' stmt	*/
			/* (though its enclosing proc is!).			*/
			/*	If the stmt is labelled, we must re-install a	*/
			/* label record for it, as the current stmt actually	*/
			/* belongs to the enclosing unit, which has just become	*/
			/* the current unit ('*this_unitp').			*/
			/*------------------------------------------------------*/
			remove_unit_record ();	/* this deletes all labels, and        */
						/* zeroes 'stmt_lp' & 'stmt_newlabel'! */
			if (stmt_label)
				lookup_or_install_label (stmt_label, SET_LABEL);
		}
		/*}}}*/
	}

	if (incl_proc  &&  (proc = this_unitp->attrs & UNIT_PROC_TYPE)) {
		errmsg ("missing 'END' statement for %s %q",
			 str_proc_type (proc),
			 this_unitp->sp ? this_unitp->sp->name : 0);
		close_current_unit ();
	}
}
/*}}}*/

/*}}}*/
